//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published 
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

// Author: Asbjrn Heid <lordcrc@gmail.com>

#region UICode
#endregion

class MathUtil
{
    private MathUtil()
    {
    }
    
    public static float Lerp(float t, float v_min, float v_max)
    {
        return (1.0f - t) * v_min + t * v_max;
    }    
}

class Gamma
{
    public Gamma(double g)
    {
        toLinear = new float[256];
        for (int i = 0; i < 256; i++)
        {
            toLinear[i] = (float)Math.Pow(i / 255.0, g);
        }
        fromLinear = new byte[65536];
        for (int i = 0; i < 65536; i++)
        {
            fromLinear[i] = (byte)(255 * Math.Pow(i / 65535.0, 1.0 / g));
        }
    }
    
    public byte FromLinear(float v)
    {
        int x = Math.Max(0, Math.Min((int)(v * 65535), 65535));
        return fromLinear[x];
    }
    
    public float ToLinear(byte v)
    {
        return toLinear[v];
    }
        
    private float[] toLinear;
    private byte[] fromLinear;
}

void Render(Surface dst, Surface src, Rectangle rect)
{
    ColorBgra PrimaryColor = (ColorBgra)EnvironmentParameters.PrimaryColor;

    Gamma g = new Gamma(2.2);
    
    float wpR = g.ToLinear(PrimaryColor.R);
    float wpG = g.ToLinear(PrimaryColor.G);
    float wpB = g.ToLinear(PrimaryColor.B);
    
    if (wpR <= 0 && wpG <= 0 && wpB <= 0)
    {
        wpR = 1;
        wpG = 1;
        wpB = 1;
    }
    
    float wpY = (wpR + wpG) / 2;
    float wpC = (wpG + wpB) / 2;
    float wpM = (wpR + wpB) / 2;
    
    // filter weights
    float wC = wpR / (wpR + wpG + wpB);
    float wM = wpG / (wpR + wpG + wpB);
    float wY = wpB / (wpR + wpG + wpB);
    float wR = wpC / (wpC + wpM + wpY);
    float wG = wpM / (wpC + wpM + wpY);
    float wB = wpY / (wpC + wpM + wpY);

    // filter coefficients
    float fR = wM + wY;
    float fG = wY + wC;
    float fB = wC + wM;
    float fC = wG + wB;
    float fM = wB + wR;
    float fY = wR + wG;
    
    // rgb vs cmy weight
    float v;
    if (wpR < wpG && wpR < wpB)
    {
        // cyan dominance
        v = 1.0f - (Math.Max(wpG, wpB) - wpC) / wpC;
    } else if (wpG < wpB) 
    {
        // magenta dominance
        v = 1.0f - (Math.Max(wpR, wpB) - wpM) / wpM;
    } else 
    {
        // yellow dominance
        v = 1.0f - (Math.Max(wpR, wpG) - wpY) / wpY;
    }
    
    ColorBgra c;
    for (int y = rect.Top; y < rect.Bottom; y++)
    {
        for (int x = rect.Left; x < rect.Right; x++)
        {
            c = src[x,y];
            
            float cR = g.ToLinear(c.R);
            float cG = g.ToLinear(c.G);
            float cB = g.ToLinear(c.B);
            float cY = (cR + cG) * 0.5f;
            float cC = (cG + cB) * 0.5f;
            float cM = (cR + cB) * 0.5f;
            
            float cL = 0.299f * cR + 0.587f * cG + 0.114f * cB;
            
            if (cL <= 0.0f)
                continue;
            
            // filter
            cR *= fR;
            cG *= fG;
            cB *= fB;
            cY *= fY;
            cC *= fC;
            cM *= fM;
           
            cR = MathUtil.Lerp(v, cR, (cY - cC + cM));
            cG = MathUtil.Lerp(v, cG, (cC - cM + cY));
            cB = MathUtil.Lerp(v, cB, (cM - cY + cC));                      
                        
            float cFL = 0.299f * cR + 0.587f * cG + 0.114f * cB;

            // recover brightness 
            float Lfactor = cL / cFL;
            
            cR *= Lfactor;
            cG *= Lfactor;
            cB *= Lfactor;
            
            //cR = cG = cB = v;
            
            dst[x,y] = ColorBgra.FromBgra(g.FromLinear(cB), g.FromLinear(cG), g.FromLinear(cR), 255);
        }
    }
}
